/*
 * PosixRealTimeExecutionContext.cpp
 * Masaharu Shimizu@fuRo
 * 08/9/1
 *
 */


#include "PosixRealTimeExecutionContext.h"
#include <rtm/ECFactory.h>

#include <iostream>

namespace RTC {

#define MLOCK

PosixRealTimeExecutionContext::PosixRealTimeExecutionContext() :
	PeriodicExecutionContext() {
}

PosixRealTimeExecutionContext::~PosixRealTimeExecutionContext() {
}

void PosixRealTimeExecutionContext::timespec_add_ns(struct timespec *a, unsigned int ns)
{
	ns += a->tv_nsec;
	while(ns >= NSEC_PER_SEC) {
		ns -= NSEC_PER_SEC;
		a->tv_sec++;
	}
	a->tv_nsec = ns;
};

int PosixRealTimeExecutionContext::timespec_compare(const struct timespec *lhs, const struct timespec *rhs) {
	if (lhs->tv_sec < rhs->tv_sec)
		return -1;
	if (lhs->tv_sec > rhs->tv_sec)
		return 1;
	return lhs->tv_nsec - rhs->tv_nsec;
};

int PosixRealTimeExecutionContext::svc(void) {
	struct timespec ts_period, ts_now;
	struct sched_param param;

	//mlockallを用いた方が性能がよいとされるが、Ubuntuの場合ルート権限が必要になるため
	//場合によって無効化するのもあり
#ifdef MLOCK
	if(mlockall(MCL_CURRENT | MCL_FUTURE))
		std::cerr << "fatal error: mlockall" << std::endl;
	mallopt(M_TRIM_THRESHOLD, -1);
	mallopt(M_MMAP_MAX, 0);
#endif

	param.sched_priority = sched_get_priority_max(SCHED_FIFO)-1;
	if (sched_setscheduler(0, SCHED_FIFO, &param) > 0)
		std::cerr << "fatal error: sched_setsheduler" << std::endl;

	clock_gettime(CLOCK_MONOTONIC, &ts_period);
	do {
		clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &ts_period, 0);
		std::for_each(m_comps.begin(), m_comps.end(), invoke_worker());

		timespec_add_ns(&ts_period, m_usec * 1000);

		clock_gettime(CLOCK_MONOTONIC, &ts_now);
		if (timespec_compare(&ts_period, &ts_now) < 0) {
			ts_period = ts_now;
			timespec_add_ns(&ts_period, m_usec * 1000);
		}
	} while (m_running);

	return 0;
}
};

extern "C" {
void PosixRealTimeExecutionContextInit(RTC::Manager* manager) {
	RTC::Manager::instance().registerECFactory("PosixRealTimeExecutionContext", RTC::ECCreate<RTC::PosixRealTimeExecutionContext>,
	RTC::ECDelete<RTC::PosixRealTimeExecutionContext>);

}
};
